import operator
from typing import Union, Iterable, Optional
from contextlib import suppress

from numba import njit, prange
from numba.core import types as nbtypes, cgutils
from numba.extending import (
    typeof_impl,
    models,
    make_attribute_wrapper,
    register_model,
    box,
    unbox,
    NativeValue,
    type_callable,
    lower_builtin,
    overload,
    overload_attribute,
)

from numpy import ndarray
import numpy as np

from sigmaepsilon.core.typing import issequence
from sigmaepsilon.math import minmax
from sigmaepsilon.math.linalg import Vector, FrameLike

from .frame import CartesianFrame
from .point import Point
from ..utils.space import index_of_closest_point, index_of_furthest_point

__cache = True


__all__ = ["PointCloud"]


VectorLike = Union[Point, Vector, ndarray]


@njit(nogil=True, parallel=True, cache=__cache)
def show_coords(dcm: ndarray, coords: ndarray) -> ndarray:
    res = np.zeros_like(coords)
    for i in prange(coords.shape[0]):
        res[i] = dcm @ coords[i, :]
    return res


def dcoords(coords: ndarray, v: ndarray) -> ndarray:
    res = np.zeros_like(coords)
    if len(res.shape) == 1:
        with suppress(IndexError):
            res[0] = v[0]
            res[1] = v[1]
            res[2] = v[2]
    elif len(res.shape) == 2:
        with suppress(IndexError):
            res[:, 0] = v[0]
            res[:, 1] = v[1]
            res[:, 2] = v[2]
    else:  # pragma: no cover
        raise ValueError("Only 1 and 2 dimensional arrays are supported here.")
    return res


class PointCloud(Vector):
    """
    A numba-jittable class to support calculations related to points
    in Euclidean space.

    Parameters
    ----------
    frame: FrameLike or numpy.ndarray, Optional
        A numpy array representing coordinate axes of a reference frame.
        Default is None.
    inds: numpy.ndarray, Optional
        An 1d integer array specifying point indices. Default is None.

    Examples
    --------
    Collect the points of a simple triangulation and get the center:

    >>> from sigmaepsilon.mesh.space import PointCloud
    >>> from sigmaepsilon.mesh.triang import triangulate
    >>> coords, *_ = triangulate(size=(800, 600), shape=(10, 10))
    >>> coords = PointCloud(coords)
    >>> coords.center()
    array([400., 300.,   0.])

    Centralize and get center again:

    >>> coords = coords.centralize()
    >>> center = coords.center()  # array([0., 0., 0.])

    Move the points in the global frame:

    >>> coords = coords.move(np.array([1., 0., 0.]))
    >>> center = coords.center()  # array([1., 0., 0.])

    Rotate the points with 90 degrees around global Z.
    Before we do so, let check the boundaries:

    >>> coords.x().min(), coords.x().max()
    (Array(-399.), Array(401.))

    >>> coords.y().min(), coords.y().max()
    (Array(-300.), Array(300.))

    Now centralize wrt. the global frame, rotate and check
    the boundaries again:

    >>> coords = coords.rotate('Body', [0, 0, np.pi], 'XYZ')
    >>> center = coords.center()  # [1., 0., 0.]

    >>> coords.x().min(), coords.x().max()
    (Array(-401.), Array(399.))

    >>> coords.y().min(), coords.y().max()
    (Array(-300.), Array(300.))

    The object keeps track of indices after slicing, always
    referring to the top level array:

    >>> coords[10:50][[1, 2, 10]].inds
    array([11, 12, 20])

    The instances are available in Numba-jitted functions, with
    the coordinates and indices available as 'data' and 'inds':

    >>> from numba import jit
    >>> @jit(nopython=True)
    ... def foo(arr): return arr.data, arr.inds
    >>> c = np.array([[0, 0, 0], [0, 0, 1.], [0, 0, 0]])
    >>> COORD = PointCloud(c, inds=np.array([0, 1, 2, 3]))
    >>> data, inds = foo(COORD)
    """

    _frame_cls_ = CartesianFrame

    def __init__(
        self,
        *args,
        frame: Optional[Union[CartesianFrame, FrameLike, None]] = None,
        inds: Optional[Union[ndarray, None]] = None,
        **kwargs,
    ):
        if frame is None:
            if len(args) > 0 and isinstance(args[0], ndarray):
                frame = self._frame_cls_(dim=args[0].shape[-1])
        super().__init__(*args, frame=frame, **kwargs)
        self.inds = inds if inds is None else np.array(inds, dtype=int)

    def __getitem__(self, key):
        inds = None
        key = (key,) if not isinstance(key, tuple) else key
        if isinstance(key[0], slice):
            slc = key[0]
            start, stop, step = slc.start, slc.stop, slc.step
            start = 0 if start == None else start
            step = 1 if step == None else step
            stop = self.shape[0] if stop == None else stop
            inds = list(range(start, stop, step))
        elif issequence(key[0]):
            inds = key[0]
        elif isinstance(key[0], int):
            inds = [
                key[0],
            ]
        if inds is not None and self.inds is not None:
            inds = self.inds[inds]
        arr = self._array.__getitem__(key)
        return PointCloud(arr, frame=self.frame, inds=inds)

    @property
    def frame(self) -> Union[CartesianFrame, FrameLike]:
        """
        Returns the frame the points are embedded in.
        """
        return self._frame

    @frame.setter
    def frame(self, target: FrameLike):
        """
        Sets the frame. This changes the frame itself and results
        in a transformation of coordinates.

        Parameters
        ----------
        target: FrameLike
            A target frame of reference.
        """
        if isinstance(target, FrameLike):
            self._array = self.show(target)
            self._frame = target
        else:
            raise TypeError("Value must be a {} instance".format(FrameLike))

    @property
    def id(self) -> ndarray:
        """
        Returns the indices of the points.
        """
        return self.inds

    def x(self, target: FrameLike = None) -> ndarray:
        """Returns the `x` coordinates."""
        arr = self.show(target)
        return arr[:, 0] if len(self.shape) > 1 else arr[0]

    def y(self, target: FrameLike = None) -> ndarray:
        """Returns the `y` coordinates."""
        arr = self.show(target)
        return arr[:, 1] if len(self.shape) > 1 else arr[1]

    def z(self, target: FrameLike = None) -> ndarray:
        """Returns the `z` coordinates."""
        arr = self.show(target)
        return arr[:, 2] if len(self.shape) > 1 else arr[2]

    def bounds(self, target: FrameLike = None) -> ndarray:
        """
        Returns the bounds of the pointcloud as a numpy array with
        a shape of (N, 2), where N is 2 for 2d problems and 3 for 3d
        ones.
        """
        arr = self.show(target)
        dim = arr.shape[1]
        res = np.zeros((dim, 2))
        res[0] = minmax(arr[:, 0])
        res[1] = minmax(arr[:, 1])
        if dim > 2:
            res[2] = minmax(arr[:, 2])
        return res

    def center(self, target: FrameLike = None) -> ndarray:
        """
        Returns the center of the points in a specified frame,
        or the root frame if there is no target provided.

        Parameters
        ----------
        target: ReferenceFrame, Optional
            A frame of reference. Default is None.

        Returns
        -------
        numpy.ndarray
            A numpy array.
        """
        arr = self.show(target)

        def mean(i: int) -> float:
            return np.mean(arr[:, i])

        return np.array(list(map(mean, range(self.shape[1]))))

    def index_of_closest(self, p: VectorLike, frame: FrameLike = None) -> int:
        """
        Returns the index of the point being closest to `p`.

        Parameters
        ----------
        p: Vector or Array, Optional
            Vectors or coordinates of one or more points. If provided as
            an array, the `frame` argument can be used to specify the
            parent frame in which the coordinates are to be understood.
        frame: ReferenceFrame, Optional
            A frame in which the input is defined if it is not a Vector.
            Default is None.

        Returns
        -------
        int
        """
        if not isinstance(p, Vector):
            p = np.array(p)
            if frame is None:
                frame = self._frame_cls_(dim=p.shape[-1])
            p = Vector(p, frame=frame)
        return index_of_closest_point(self.show(), p.show())

    def index_of_furthest(self, p: VectorLike, frame: FrameLike = None) -> int:
        """
        Returns the index of the point being furthest from `p`.

        Parameters
        ----------
        p: Vector or Array, Optional
            Vectors or coordinates of one or more points. If provided as
            an array, the `frame` argument can be used to specify the
            parent frame in which the coordinates are to be understood.
        frame: ReferenceFrame, Optional
            A frame in which the input is defined if it is not a Vector.
            Default is None.

        Returns
        -------
        int
        """
        if not isinstance(p, Vector):
            p = Vector(p, frame=frame)
        return index_of_furthest_point(self.show(), p.show())

    def closest(self, p: VectorLike, frame: FrameLike = None) -> Point:
        """
        Returns the point being closest to `p`.

        Parameters
        ----------
        p: Vector or Array, Optional
            Vectors or coordinates of one or more points. If provided as
            an array, the `frame` argument can be used to specify the
            parent frame in which the coordinates are to be understood.
        frame: ReferenceFrame, Optional
            A frame in which the input is defined if it is not a Vector.
            Default is None.

        Returns
        -------
        `~sigmaepsilon.mesh.space.point.Point`
        """
        id = self.index_of_closest(p, frame)
        arr = self._array[id, :]
        if isinstance(self.inds, np.ndarray):
            gid = self.inds[id]
        else:
            gid = id
        if isinstance(id, int):
            return Point(arr, frame=self.frame, id=id, gid=gid)
        else:
            return PointCloud(arr, frame=self.frame, inds=id)

    def furthest(self, p: VectorLike, frame: FrameLike = None) -> Point:
        """
        Returns the point being closest to `p`.

        Parameters
        ----------
        p: Vector or Array, Optional
            Vectors or coordinates of one or more points. If provided as
            an array, the `frame` argument can be used to specify the
            parent frame in which the coordinates are to be understood.
        frame: ReferenceFrame, Optional
            A frame in which the input is defined if it is not a Vector.
            Default is None.

        Returns
        -------
        `~sigmaepsilon.mesh.space.point.Point`
        """
        id = self.index_of_furthest(p, frame)
        arr = self._array[id, :]
        if isinstance(self.inds, np.ndarray):
            gid = self.inds[id]
        else:
            gid = id
        if isinstance(id, int):
            return Point(arr, frame=self.frame, id=id, gid=gid)
        else:
            return PointCloud(arr, frame=self.frame, inds=id)

    def show(self, target: FrameLike = None, *args, **kwargs) -> ndarray:
        """
        Returns the coordinates of the points in a specified frame,
        or the root frame if there is no target provided.

        Parameters
        ----------
        target: ReferenceFrame, Optional
            A frame of reference. Default is None.

        Notes
        -----
        This function returns the coordinates of the points in a target
        frame, but does not make any changes to the points themselves.
        If you want to change the frame of the pointcloud, reset the
        frame of the object by setting the `frame` property.

        See Also
        --------
        :func:`~sigmaepsilon.mesh.space.pointcloud.PointCloud.frame`

        Returns
        -------
        numpy.ndarray
            The coordinates in the desired frame.
        """
        # passing unexpected arguments is necessary here because the
        # function might ocassionally be called from super()
        x = super().show(target, *args, **kwargs)
        frame = self.frame
        if isinstance(frame, CartesianFrame):
            buf = x + dcoords(x, self.frame.relative_origo(target))
        else:
            buf = x
        return self._array_cls_(shape=buf.shape, buffer=buf, dtype=buf.dtype)

    def move(self, v: VectorLike, frame: FrameLike = None) -> "PointCloud":
        """
        Moves the points wrt. to a specified frame, or the root
        frame if there is no target provided. Returns the object
        for continuation.

        Parameters
        ----------
        v: Vector or Array, Optional
            An array of a vector. If provided as an array, the `frame`
            argument can be used to specify the parent frame in which the
            motion is tp be understood.
        frame: ReferenceFrame, Optional
            A frame in which the input is defined if it is not a Vector.
            Default is None.

        Returns
        -------
        PointCloud
            The object the function is called on.

        Examples
        --------
        Collect the points of a simple triangulation and get the center:

        >>> from sigmaepsilon.mesh import triangulate
        >>> coords, *_ = triangulate(size=(800, 600), shape=(10, 10))
        >>> coords = PointCloud(coords)
        >>> coords.center()
        array([400., 300.,   0.])

        Move the points and get the center again:

        >>> d = np.array([0., 1., 0.])
        >>> coords = coords.move(d).move(d)
        >>> coords.center()
        array([400., 302.,   0.])
        """
        if not isinstance(v, Vector) and frame:
            v = Vector(v, frame=frame)
            arr = v.show(self.frame)
        else:
            arr = v
        self._array += dcoords(self._array, arr)
        return self

    def centralize(
        self, target: FrameLike = None, axes: Iterable = None
    ) -> "PointCloud":
        """
        Centralizes the coordinates wrt. to a specified frame,
        or the root frame if there is no target provided.

        Returns the object for continuation.

        Parameters
        ----------
        target: ReferenceFrame, Optional
            A frame of reference. Default is None.
        axes: Iterable, Optional
            The axes on which centralization is to be performed. A `None` value
            means all axes. Default is None.

        Returns
        -------
        PointCloud
            The object the function is called on.
        """
        center = self.center(target)
        d = np.zeros_like(center, dtype=float)
        if not isinstance(axes, Iterable):
            axes = list(range(len(center)))
        d[axes] = center[axes]
        return self.move(-d, target)

    def rotate(self, *args, **kwargs) -> "PointCloud":
        """
        Applies a transformation to the coordinates in-place. All arguments
        are passed to `ReferenceFrame.orient_new`, see its docs to know more.

        Returns the object for continuation.

        Parameters
        ----------
        *args: tuple,
            The first positional argument can be a ReferenceFrame object.
            If it is not, all positional and keyword arguments are forwarded
            to `ReferenceFrame.orient_new`.

        Returns
        -------
        PointCloud
            The object the function is called on.

        Examples
        --------
        To apply a 90 degree rotation about the Z axis:

        >>> from sigmaepsilon.mesh.space import PointCloud
        >>> from sigmaepsilon.mesh import triangulate
        >>> coords, *_ = triangulate(size=(800, 600), shape=(10, 10))
        >>> points = PointCloud(coords)
        >>> rotated_pointcloud = points.rotate('Space', [0, 0, np.pi/2], 'XYZ')
        """
        if isinstance(args[0], FrameLike):
            self.orient(dcm=args[0].dcm())
            return self
        else:
            target = self.frame.orient_new(*args, **kwargs)
            return self.rotate(target)

    def idsort(self) -> ndarray:
        """
        Returns the indices that would sort the array according to
        their indices.
        """
        return np.argsort(self.inds)

    def sort_indices(self) -> "PointCloud":
        """
        Sorts the points according to their indices and returns the
        object.
        """
        s = self.idsort()
        self._array = self._array[s]
        self.inds = self.inds[s]
        return self

    def __repr__(self):
        return f"PointCloud({self._array})"

    def __str__(self):
        return f"PointCloud({self._array})"


class PointCloudType(nbtypes.Type):  # pragma: no cover
    """Numba type."""

    def __init__(self, datatype, indstype=nbtypes.NoneType):
        self.data = datatype
        self.inds = indstype
        super(PointCloudType, self).__init__(name="PointCloud")


make_attribute_wrapper(PointCloudType, "data", "data")
make_attribute_wrapper(PointCloudType, "inds", "inds")


@overload_attribute(PointCloudType, "x")
def attr_x(arr):  # pragma: no cover
    def get(arr):
        return arr.data[:, 0]

    return get


@overload_attribute(PointCloudType, "y")
def attr_y(arr):  # pragma: no cover
    def get(arr):
        return arr.data[:, 1]

    return get


@overload_attribute(PointCloudType, "z")
def attr_z(arr):  # pragma: no cover
    def get(arr):
        return arr.data[:, 2]

    return get


@typeof_impl.register(PointCloud)
def type_of_impl(val, context):  # pragma: no cover
    """`val` is the Python object being typed"""
    datatype = typeof_impl(val._array, context)
    indstype = typeof_impl(val.inds, context)
    return PointCloudType(datatype, indstype)


@type_callable(PointCloud)
def type_of_callable(context):  # pragma: no cover
    def typer(data, inds=None):
        datatype = typeof_impl(data, context)
        indstype = typeof_impl(inds, context) if inds is not None else nbtypes.NoneType
        return PointCloudType(datatype, indstype)

    return typer


@register_model(PointCloudType)
class StructModel(models.StructModel):  # pragma: no cover
    """Data model for nopython mode."""

    def __init__(self, dmm, fe_type):
        """
        fe_type is `PointCloudType`
        """
        members = [
            ("data", fe_type.data),
            ("inds", fe_type.inds),
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


@overload(operator.getitem)
def overload_getitem(obj, idx):  # pragma: no cover
    if isinstance(obj, PointCloudType):

        def dummy_getitem_impl(obj, idx):
            return obj.data[idx]

        return dummy_getitem_impl


@lower_builtin(PointCloud, nbtypes.Array)
def lower_type(context, builder, sig, args):  # pragma: no cover
    typ = sig.return_type
    data, inds = args
    obj = cgutils.create_struct_proxy(typ)(context, builder)
    obj.data = data
    obj.inds = inds
    return obj._getvalue()


@unbox(PointCloudType)
def unbox_type(typ, obj, c):  # pragma: no cover
    """Convert a python object to a numba-native structure."""
    data_obj = c.pyapi.object_getattr_string(obj, "_array")
    inds_obj = c.pyapi.object_getattr_string(obj, "inds")
    native_obj = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    native_obj.data = c.unbox(typ.data, data_obj).value
    native_obj.inds = c.unbox(typ.inds, inds_obj).value
    c.pyapi.decref(data_obj)
    c.pyapi.decref(inds_obj)
    is_error = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(native_obj._getvalue(), is_error=is_error)


@box(PointCloudType)
def box_type(typ, val, c):  # pragma: no cover
    """Convert a numba-native structure to a python object."""
    native_obj = cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(PointCloud))
    data_obj = c.box(typ.data, native_obj.data)
    inds_obj = c.box(typ.inds, native_obj.inds)
    python_obj = c.pyapi.call_function_objargs(class_obj, (data_obj, inds_obj))
    c.pyapi.decref(data_obj)
    c.pyapi.decref(inds_obj)
    return python_obj



import numpy as np
from typing import Union, Iterable, Optional
from copy import deepcopy as dcopy

import numpy as np
from numpy import ndarray

from sigmaepsilon.math.linalg import CartesianFrame as Frame, FrameLike, Vector
from sigmaepsilon.math.linalg.vector import Vector

__all__ = ["CartesianFrame"]


VectorLike = Union[Vector, ndarray]


class CartesianFrame(Frame):
    """
    A field-specific reference frame to be used in problems related to
    Euclidean geometry.

    It builds on top of :class:`~sigmaepsilon.math.linalg.meta.FrameLike` from
    `sigmaepsilon.math`, but adds the contept of 'origo', and some other applications
    related to the field.

    See Also
    --------
    :class:`~sigmaepsilon.math.linalg.meta.FrameLike`

    Parameters
    ----------
    axes: numpy.ndarray, Optional.
        2d numpy array of floats specifying cartesian reference frames.
        Dafault is None.
    dim: int, Optional
        Dimension of the mesh. Deafult is 3.
    origo: numpy.ndarray, Optional.
        The origo of the mesh. Default is the zero vector.

    Note
    ----
    See the documentation of :class:`~sigmaepsilon.math.linalg.FrameLike` for more control over
    object creation. However, if your problem not very extreme in some sense,
    you are probably good to goo only by following the examples.

    Example
    -------
    Define a standard Cartesian frame and rotate it around axis 'Z'
    with an amount of 180 degrees:

    >>> from sigmaepsilon.mesh.space import CartesianFrame
    >>> import numpy as np

    >>> A = CartesianFrame(dim=3)
    >>> B = A.orient_new('Space', [0, 0, np.pi], 'XYZ')

    To create a third frame that rotates from B the way B rotates from A, we
    can do

    >>> A = CartesianFrame(dim=3)
    >>> C = A.orient_new('Space', [0, 0, 2*np.pi], 'XYZ')

    Then, the *DCM from A to B* , that is :math:`^{A}\mathbf{R}^{B}` would be

    >>> A_R_B = B.dcm(source=A)

    or equivalently

    >>> A_R_B = A.dcm(target=A)
    """

    def __init__(
        self, axes: ndarray = None, *args, dim: int = 3, origo: ndarray = None, **kwargs
    ):
        axes = np.eye(dim) if axes is None else axes
        super().__init__(axes, *args, **kwargs)
        if origo is not None:
            if not isinstance(origo, ndarray) and isinstance(origo, Iterable):
                origo = np.array(origo, dtype=float)

            if isinstance(origo, ndarray):
                if not len(origo.shape) == 1:
                    raise ValueError("'origo' must be a 1d iterable!")
            else:
                raise TypeError("'origo' must be a NumPy array or an iterable!")

        self._origo = origo

    @property
    def origo(self):
        if not isinstance(self._origo, ndarray):
            self._origo = np.zeros(len(self.axes))
        return self._origo

    @origo.setter
    def origo(self, value: Iterable):
        value = np.array(value).astype(float)
        if not isinstance(self._origo, ndarray):
            self._origo = np.zeros(len(self.axes))
        if value.shape == self._array.shape:
            if self._weakrefs and len(self._weakrefs) > 0:
                for v in self._weakrefs.values():
                    v.array -= value
            self._array = value
        else:
            raise ValueError("Mismatch in data dimensinons!")

    def relative_origo(self, target: FrameLike = None) -> ndarray:
        """
        Returns the origo of the current frame in ambient space
        or with respect to another frame.

        Parameters
        ----------
        target: FrameLike, Optional
            A frame in which we want to get the origo of the current frame.
            A None value returns the origo of the current frame with respect
            to the root. Default is None.

        Returns
        -------
        numpy.ndarray
            A vector defined in ambient space, or the specified frame.

        Examples
        --------
        Define a standard Cartesian frame and rotate it around axis 'Z'
        with an amount of 180 degrees:

        >>> from sigmaepsilon.mesh.space import CartesianFrame
        >>> import numpy as np

        >>> A = CartesianFrame()
        >>> B = A.orient_new('Space', [0, 0, 45*np.pi/180],  'XYZ')

        To get the origin of frame B:

        >>> origo = B.relative_origo()  # array([0., 0., 0.])

        Move frame B (the motion is defined locally) and print the
        new point of origin with respect to A:

        >>> B = B.move(Vector([1, 0, 0], frame=B))
        >>> origo = B.relative_origo(A)  # [0.7071, 0.7071, 0.]

        Of course, the point of origin of a frame with respect to itself
        must be a zero vector:

        >>> origo = B.relative_origo(B)  # [0., 0., 0.]

        Providing with no arguments returns the distance of origin with
        respect to the root frame:

        >>> origo = B.relative_origo()  # same as B.relative_origo(B.root), [0.7071, 0.7071, 0.]
        """
        if not isinstance(self._origo, ndarray):
            self._origo = np.zeros(len(self.axes))

        if target is None:
            return Vector(self._origo).show()
        else:
            s = self.relative_origo()
            if isinstance(target, CartesianFrame):
                t = target.relative_origo()
            else:
                t = np.zeros_like(s)
            return Vector(s - t).show(target)

    def rotate(self, *args, **kwargs) -> "CartesianFrame":
        return super().rotate(*args, **kwargs)

    def move(self, d: VectorLike, frame: FrameLike = None) -> "CartesianFrame":
        """
        Moves the frame by shifting its origo.

        Parameters
        ----------
        d: VectorLike
            :class:`Vector` or :class:`Array`, the amount of the motion.
        frame: FrameLike, Optional
            A frame in which the input is defined if it is not a Vector.
            Default is None, which assumes the root frame.

        Returns
        -------
        CartesianFrame
            The object the function is called on.

        Examples
        --------
        >>> from sigmaepsilon.mesh.space import CartesianFrame
        >>> import numpy as np

        >>> A = CartesianFrame()
        >>> v = Vector([1., 0., 0.], frame=A)
        >>> B = A.fork('Space', [0, 0, 45*np.pi/180], 'XYZ').move(v)

        Move the frame locally with the same amount

        >>> B.move(v.array, frame=B)
        array([[ 0.70710678,  0.70710678,  0.        ],
               [-0.70710678,  0.70710678,  0.        ],
               [ 0.        ,  0.        ,  1.        ]])
        """
        if not isinstance(d, Vector):
            if not isinstance(d, ndarray):
                d = np.array(d)
            d = Vector(d, frame=frame)
        if self._origo is None:
            self._origo = np.zeros(len(self.axes))
        self._origo += d.show()
        return self

    def fork(self, *args, **kwargs) -> "CartesianFrame":
        """
        Alias for `orient_new`.
        """
        return self.orient_new(*args, **kwargs)

    def copy(self, deep: bool = False, name: str = None) -> "CartesianFrame":
        """
        Returns a shallow or deep copy of this object, depending of the
        argument `deepcopy` (default is False).

        Parameters
        ----------
        deep: bool, Optional
            If True, a deep copy is returned. Default is False.
        name: str, Optional
            The name of the copy. Default is None.
        """
        if deep:
            return self.__class__(dcopy(self.axes), origo=dcopy(self.origo), name=name)
        else:
            return self.__class__(self.axes, origo=self.origo, name=name)

    def deepcopy(self, name: str = None) -> "CartesianFrame":
        """
        Returns a deep copy of the instance.

        Parameters
        ----------
        name: str, Optional
            The name of the copy. Default is None.
        """
        return self.copy(deep=True, name=name)



from typing import Tuple, List

import numpy as np
from numpy import ndarray
from sympy import symbols

from ..geometry import PolyCellGeometry3d
from ..data.polycell import PolyCell
from ..utils.numint import Gauss_Legendre_Wedge_3x3
from ..utils.cells.w18 import monoms_W18
from ..utils.topology import compose_trmap
from .w6 import W6


class W18(PolyCell):
    """
    Polyhedra class for 18-noded biquadratic wedges.
    """

    label = "W18"

    class Geometry(PolyCellGeometry3d):
        number_of_nodes = 18
        vtk_cell_id = 32
        monomial_evaluator = monoms_W18
        quadrature = {
            "full": Gauss_Legendre_Wedge_3x3,
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s, t = symbols("r s t", real=True)
            monoms = [
                1,
                r,
                s,
                r**2,
                s**2,
                r * s,
                t,
                t * r,
                t * s,
                t * r**2,
                t * s**2,
                t * r * s,
                t**2,
                t**2 * r,
                t**2 * s,
                t**2 * r**2,
                t**2 * s**2,
                t**2 * r * s,
            ]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            return np.array(
                [
                    [-1 / 3, -1 / 3, -1.0],
                    [2 / 3, -1 / 3, -1.0],
                    [-1 / 3, 2 / 3, -1.0],
                    [-1 / 3, -1 / 3, 1.0],
                    [2 / 3, -1 / 3, 1.0],
                    [-1 / 3, 2 / 3, 1.0],
                    [1 / 6, -1 / 3, -1.0],
                    [1 / 6, 1 / 6, -1.0],
                    [-1 / 3, 1 / 6, -1.0],
                    [1 / 6, -1 / 3, 1.0],
                    [1 / 6, 1 / 6, 1.0],
                    [-1 / 3, 1 / 6, 1.0],
                    [-1 / 3, -1 / 3, 0.0],
                    [2 / 3, -1 / 3, 0.0],
                    [-1 / 3, 2 / 3, 0.0],
                    [1 / 6, -1 / 3, 0.0],
                    [1 / 6, 1 / 6, 0.0],
                    [-1 / 3, 1 / 6, 0.0],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            return np.array([[0.0, 0.0, 0.0]], dtype=float)

        @classmethod
        def tetmap(cls) -> ndarray:
            w18_to_w6 = np.array(
                [
                    [15, 13, 16, 9, 4, 10],
                    [17, 16, 14, 11, 10, 5],
                    [17, 15, 16, 11, 9, 10],
                    [12, 15, 17, 3, 9, 11],
                    [6, 1, 7, 15, 13, 16],
                    [8, 6, 7, 17, 15, 16],
                    [8, 7, 2, 17, 16, 14],
                    [8, 0, 6, 17, 12, 15],
                ],
                dtype=int,
            )
            w6_to_tet4 = W6.Geometry.tetmap()
            return compose_trmap(w18_to_w6, w6_to_tet4)



import numpy as np
from contextlib import suppress

from .polydata import PolyData
from ..cells import L2, L3

from ..config import __hasplotly__

if __hasplotly__:
    with suppress(ImportError):
        from sigmaepsilon.plotting.plotly import plot_lines_3d


__all__ = ["LineData"]


class LineData(PolyData):
    """Data class for 1d cells."""

    def _init_config_(self):
        super()._init_config_()
        key = self.__class__._pv_config_key_
        self.config[key]["color"] = "k"
        self.config[key]["line_width"] = 10
        self.config[key]["render_lines_as_tubes"] = True

    def __plot_plotly__(self, *, scalars=None, fig=None, **kwargs):
        """
        Plots collections of lines and data provided on the nodes using `plotly`.

        Returns the figure object.
        """
        coords = self.coords()
        topo = self.topology()
        kwargs.update(dict(scalars=scalars, fig=fig))
        return plot_lines_3d(coords, topo, **kwargs)

    def plot(self, *args, scalars=None, backend="plotly", scalar_labels=None, **kwargs):
        """
        Plots the mesh with or without data, using multiple possible backends.

        Parameters
        ----------
        scalars: numpy.ndarray, Optional
            Stacked nodal information as an 1d or 2d NumPy array. Default is None.
        backend: str, Optional
            The backend to use. Possible options are `plotly` and `vtk`
            (`vtk` is being managed through `PyVista`) at the moment.
            Default is 'plotly'.
        scalar_labels: Iterable, Optional
            Labels for the datasets provided with 'scalars'. Default is None.
        """
        if backend == "vtk":
            return self.pvplot(
                *args, scalars=scalars, scalar_labels=scalar_labels, **kwargs
            )
        elif backend == "plotly":
            if __hasplotly__:
                return self.__plot_plotly__(
                    *args, scalars=scalars, scalar_labels=None, **kwargs
                )
            else:
                msg = "You need to install `dewloosh.plotly` for this plot."
                raise ImportError(msg)
        else:
            msg = "No implementation for backend '{}'".format(backend)
            raise NotImplementedError(msg)



from .from_pyvista import from_pv
from .from_meshio import from_meshio
from .to_pyvista import to_pv
from .to_vtk import to_vtk
from .to_k3d import to_k3d

__all__ = ["from_pv", "from_meshio", "to_pv", "to_vtk", "to_k3d"]



import numpy as np
from numpy import ndarray

__all__ = ["trimap_Q8"]


def trimap_Q8() -> ndarray:
    return np.array(
        [
            [0, 4, 7],
            [4, 1, 5],
            [5, 2, 6],
            [6, 3, 7],
            [7, 5, 6],
            [4, 5, 7],
        ],
        dtype=int,
    )



from sigmaepsilon.core.downloads import download_file, delete_downloads
from sigmaepsilon.mesh import PolyData


__all__ = [
    "download_stand",
    "download_bunny",
    "delete_downloads",
    "download_bunny_coarse",
    "download_gt40",
    "download_badacsony",
    "download_bike_stem",
]


def _download(path: str, read: bool = False) -> str | PolyData:
    vtkpath = download_file(path)
    if read:
        return PolyData.read(vtkpath)
    else:
        return vtkpath


def download_stand(*, read: bool = False) -> str | PolyData:
    """
    Downloads a tetrahedral mesh of a stand in vtk format.

    Parameters
    ----------
    read: bool, Optional
        If False, the path of the mesh file is returned instead of a
        :class:`~sigmaepsilon.mesh.data.polydata.PolyData` object. Default is False.

    Returns
    -------
    str
        A path to a file on your filesystem.

    Example
    --------
    >>> from sigmaepsilon.mesh.downloads import download_stand
    >>> file_path = download_stand()
    """
    return _download("stand.vtk", read=read)


def download_bunny(*, tetra: bool = False, read: bool = False) -> str | PolyData:
    """
    Downloads a tetrahedral mesh of a bunny in vtk format.

    Parameters
    ----------
    tetra: bool, Optional
        If True, the returned mesh is a tetrahedral one, otherwise
        it is a surface triangulation. Default is False.
    read: bool, Optional
        If False, the path of the mesh file is returned instead of a
        :class:`~sigmaepsilon.mesh.data.polydata.PolyData` object. Default is False.

    Returns
    -------
    str
        A path to a file on your filesystem.

    Example
    --------
    >>> from sigmaepsilon.mesh.downloads import download_bunny
    >>> file_path = download_bunny()
    """
    filename = "bunny_T3.vtk" if not tetra else "bunny_TET4.vtk"
    return _download(filename, read=read)


def download_bunny_coarse(tetra: bool = False, read: bool = False) -> str | PolyData:
    """
    Downloads and optionally reads the bunny example as a vtk file.

    Parameters
    ----------
    tetra: bool, Optional
        If True, the returned mesh is a tetrahedral one, otherwise
        it is a surface triangulation. Default is False.
    read: bool, Optional
        If False, the path of the mesh file is returned instead of a
        :class:`~sigmaepsilon.mesh.data.polydata.PolyData` object. Default is False.

    Example
    -------
    >>> from sigmaepsilon.mesh.downloads import download_bunny_coarse
    >>> mesh = download_bunny_coarse(tetra=True, read=True)
    """
    filename = "bunny_T3_coarse.vtk" if not tetra else "bunny_TET4_coarse.vtk"
    return _download(filename, read=read)


def download_gt40(read: bool = False) -> str | PolyData:
    """
    Downloads and optionally reads the Gt40 example as a vtk file.

    Parameters
    ----------
    read: bool, Optional
        If False, the path of the mesh file is returned instead of a
        :class:`~sigmaepsilon.mesh.data.polydata.PolyData` object. Default is False.

    Example
    -------
    >>> from sigmaepsilon.mesh.downloads import download_gt40
    >>> mesh = download_gt40(read=True)
    """
    return _download("gt40.vtk", read=read)


def download_badacsony(read: bool = False) -> str | PolyData:
    """
    Downloads and optionally reads the badacsony example as a vtk file.

    Parameters
    ----------
    read: bool, Optional
        If False, the path of the mesh file is returned instead of a
        :class:`~sigmaepsilon.mesh.data.polydata.PolyData` object. Default is False.

    Example
    -------
    >>> from sigmaepsilon.mesh.downloads import download_badacsony
    >>> mesh = download_badacsony(read=True)
    """
    return _download("badacsony.vtk", read=read)


def download_bike_stem(read: bool = False) -> str | PolyData:
    """
    Downloads and optionally reads the bike stem example as an STL file.

    Parameters
    ----------
    read: bool, Optional
        If False, the path of the mesh file is returned instead of a
        :class:`~sigmaepsilon.mesh.data.polydata.PolyData` object. Default is False.

    Example
    -------
    >>> from sigmaepsilon.mesh.downloads import download_bike_stem
    >>> mesh = download_bike_stem(read=True)
    """
    return _download("bike_stem_nomanifold.stl", read=read)



from ...config import __hasplotly__

if not __hasplotly__:  # pragma: no cover

    def triplot_plotly(*_, **__):
        raise ImportError(
            "You need Plotly for this. Install it with 'pip install plotly'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Optional, Union

    import plotly.graph_objects as go
    from numpy import ndarray

    from sigmaepsilon.mesh.utils.topology import unique_topo_data
    from sigmaepsilon.mesh.utils.tri import edges_tri
    from .lines import scatter_lines_plotly

    def triplot_plotly(
        points: ndarray,
        triangles: ndarray,
        data: Optional[ndarray] = None,
        plot_edges: Optional[bool] = True,
        edges: Optional[Union[bool, None]] = None,
    ) -> go.Figure:
        """
        Plots a triangulation optionally with edges and attached field data in 3d.

        Parameters
        ----------
        points: numpy.ndarray
            2d float array of points.
        data: numpy.ndarray
            1d float array of scalar data over the points.
        plot_edges: bool, Optional
            If True, plots the edges of the mesh. Default is False.
        edges: numpy.ndarray, Optional
            The edges to plot. If provided, `plot_edges` is ignored. Default is None.

        Returns
        -------
        figure: :class:`plotly.graph_objects.Figure`
            The figure object.

        Example
        -------
        .. plotly::
            :include-source: True

            from sigmaepsilon.mesh.plotting import triplot_plotly
            from sigmaepsilon.mesh import grid
            from sigmaepsilon.mesh.utils.topology.tr import Q4_to_T3
            import numpy as np
            gridparams = {
                "size": (1200, 600),
                "shape": (4, 4),
                "eshape": (2, 2),
            }
            coords, topo = grid(**gridparams)
            points, triangles = Q4_to_T3(coords, topo, path="grid")
            data = np.random.rand(len(points))
            triplot_plotly(points, triangles, data, plot_edges=True)
        """
        x = points[:, 0]
        y = points[:, 1]
        z = points[:, 2]
        i = triangles[:, 0]
        j = triangles[:, 1]
        k = triangles[:, 2]

        if data is not None:
            fig = go.Figure(
                data=[
                    go.Mesh3d(
                        x=x,
                        y=y,
                        z=z,
                        i=i,
                        j=j,
                        k=k,
                        intensity=data,
                        opacity=1,
                    )
                ]
            )
        else:
            fig = go.Figure(
                data=[
                    go.Mesh3d(
                        x=x,
                        y=y,
                        z=z,
                        i=i,
                        j=j,
                        k=k,
                        opacity=1,
                    )
                ]
            )

        fig.update_layout(
            template="plotly",
            autosize=True,
            margin=dict(l=1, r=1, b=1, t=1, pad=0),
            scene=dict(
                aspectmode="data",
            ),
        )

        if edges is not None:
            plot_edges = True

        if plot_edges:
            if edges is None:
                edges, _ = unique_topo_data(edges_tri(triangles))

            scatter_lines_plotly(points, edges, fig=fig)

        return fig


__all__ = ["triplot_plotly"]



from ...config import __hasplotly__

if not __hasplotly__:  # pragma: no cover

    def plot_lines_plotly(*_, **__):
        raise ImportError(
            "You need Plotly for this. Install it with 'pip install plotly'. "
            "You may also need to restart your kernel and reload the package."
        )

    def scatter_lines_plotly(*_, **__):
        raise ImportError(
            "You need Plotly for this. Install it with 'pip install plotly'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    import plotly.graph_objects as go
    from numpy import ndarray

    from sigmaepsilon.mesh.utils import explode_mesh_bulk

    from .points import scatter_points_plotly

    def scatter_lines_plotly(
        coords: ndarray, topo: ndarray, fig: go.Figure = None
    ) -> go.Figure:
        X, _ = explode_mesh_bulk(coords, topo)
        x = X[:, 0]
        y = X[:, 1]
        z = X[:, 2]

        if fig is None:
            fig = go.Figure()

        def _stack_line_3d(i):
            scatter_cells = go.Scatter3d(
                x=x[2 * i : 2 * (i + 1)],
                y=y[2 * i : 2 * (i + 1)],
                z=z[2 * i : 2 * (i + 1)],
                mode="lines",
                line=dict(color="black", width=4),
                showlegend=False,
            )
            scatter_cells.update(hoverinfo="skip")
            fig.add_trace(scatter_cells)

        list(map(_stack_line_3d, range(topo.shape[0])))

        return fig

    def plot_lines_plotly(
        coords: ndarray,
        topo: ndarray,
        *args,
        scalars: ndarray = None,
        fig: go.Figure = None,
        marker_symbol: str = "circle",
        **kwargs,
    ) -> go.Figure:
        """
        Plots points and lines in 3d space optionally with data defined on the points.
        If data is provided, the values are shown in a tooltip when howering above a point.

        .. note:
            Currently only 2 noded linear lines are supported.

        Parameters
        ----------
        coords: numpy.ndarray
            The coordinates of the points, where the first axis runs along the points, the
            second along spatial dimensions.
        topo: numpy.ndarray
            The topology of the lines, where the first axis runs along the lines, the
            second along the nodes.
        scalars: numpy.ndarray
            The values to show in the tooltips of the points as a 1d or 2d NumPy array.
            The length of the array must equal the number of points. Default is None.
        marker_symbol: str, Optional
            The symbol to use for the points. Refer to Plotly's documentation for the
            possible options. Default is "circle".

        Example
        -------
        .. plotly::
            :include-source: True

            from sigmaepsilon.mesh.plotting import plot_lines_plotly
            from sigmaepsilon.mesh import grid
            from sigmaepsilon.mesh.utils.topology.tr import H8_to_L2
            import numpy as np
            gridparams = {
                "size": (10, 10, 10),
                "shape": (4, 4, 4),
                "eshape": "H8",
            }
            coords, topo = grid(**gridparams)
            coords, topo = H8_to_L2(coords, topo)
            data = np.random.rand(len(coords), 2)
            plot_lines_plotly(coords, topo, scalars=data, scalar_labels=["X", "Y"])
        """
        n2 = topo[:, [0, -1]].max() + 1
        _scalars = scalars[:n2] if scalars is not None else None

        if fig is None:
            fig = scatter_points_plotly(coords[:n2], *args, scalars=_scalars, **kwargs)
        else:
            scatter_points_plotly(
                coords[:n2], *args, scalars=_scalars, fig=fig, **kwargs
            )

        for i, _ in enumerate(fig.data):
            fig.data[i].marker.symbol = marker_symbol
            fig.data[i].marker.size = 5

        scatter_lines_plotly(coords, topo, fig=fig)

        fig.update_layout(
            template="plotly",
            autosize=True,
            # width=720,
            # height=250,
            margin=dict(l=1, r=1, b=1, t=1, pad=0),
            scene=dict(
                aspectmode="data",
                # xaxis = dict(nticks=5, range=[xmin - delta, xmax + delta],),
                # yaxis = dict(nticks=5, range=[ymin - delta, ymax + delta],),
                # zaxis = dict(nticks=5, range=[zmin - delta, zmax + delta],),
            ),
        )

        return fig


__all__ = ["plot_lines_plotly", "scatter_lines_plotly"]



from os.path import dirname, abspath
from importlib.metadata import metadata

from sigmaepsilon.core.config import namespace_package_name

from .space import PointCloud, CartesianFrame
from .data import PolyData, PointData, LineData, TriMesh
from .data import LineData as PolyData1d
from .utils import k_nearest_neighbours as KNN
from .topoarray import TopologyArray
from .triang import triangulate
from .grid import grid
from .cellapproximator import LagrangianCellApproximator
from .io import from_pv, from_meshio, to_vtk, to_pv, to_k3d
from .plotting import pvplot

__all__ = [
    "PointCloud",
    "CartesianFrame",
    "PolyData",
    "LineData",
    "PolyData1d",
    "PointData",
    "TriMesh",
    #
    "TopologyArray",
    "LagrangianCellApproximator",
    #
    "KNN",
    "triangulate",
    "grid",
    #
    "from_pv",
    "from_meshio",
    "to_pv",
    "to_vtk",
    "to_k3d",
    #
    "pvplot",
]

__pkg_name__ = namespace_package_name(dirname(abspath(__file__)), 10)
__pkg_metadata__ = metadata(__pkg_name__)
__version__ = __pkg_metadata__["version"]
__description__ = __pkg_metadata__["summary"]
del __pkg_metadata__



# -*- coding: utf-8 -*-
from functools import partial

from ..geometry import PolyCellGeometry1d
from ..data.polycell import PolyCell
from ..utils.cells.l2 import (
    shp_L2_multi,
    dshp_L2_multi,
    shape_function_matrix_L2_multi,
    monoms_L2,
)
from ..utils.numint import Gauss_Legendre_Line_Grid

__all__ = ["L2"]


class L2(PolyCell):
    """
    2-Node line element.
    """

    class Geometry(PolyCellGeometry1d):
        number_of_nodes = 2
        vtk_cell_id = 3
        shape_function_evaluator = shp_L2_multi
        shape_function_matrix_evaluator = shape_function_matrix_L2_multi
        shape_function_derivative_evaluator = dshp_L2_multi
        monomial_evaluator = monoms_L2
        quadrature = {
            "full": partial(Gauss_Legendre_Line_Grid, 2),
            "geometry": "full",
        }



# -*- coding: utf-8 -*-
from typing import Tuple, List

import numpy as np
from numpy import ndarray
from sympy import symbols

from ..geometry import PolyCellGeometry2d
from ..data.polycell import PolyCell
from ..utils.cells.t6 import (
    shp_T6_multi,
    dshp_T6_multi,
    shape_function_matrix_T6_multi,
    monoms_T6,
)
from ..utils.numint import Gauss_Legendre_Tri_3a
from ..utils.topology import T6_to_T3, T3_to_T6


class T6(PolyCell):
    """
    A class to handle 6-noded triangles.

    Example
    -------
    >>> from sigmaepsilon.mesh import TriMesh, CartesianFrame, PointData, triangulate
    >>> from sigmaepsilon.mesh.cells import T6 as CellData
    >>> from sigmaepsilon.mesh.utils.topology.tr import T3_to_T6
    >>> A = CartesianFrame(dim=3)
    >>> coords, topo, _ = triangulate(size=(800, 600), shape=(10, 10))
    >>> coords, topo = T3_to_T6(coords, topo)
    >>> pd = PointData(coords=coords, frame=A)
    >>> cd = CellData(topo=topo)
    >>> trimesh = TriMesh(pd, cd)
    >>> trimesh.area()
    480000.0
    """

    label = "T6"

    class Geometry(PolyCellGeometry2d):
        number_of_nodes = 6
        vtk_cell_id = 22
        shape_function_evaluator = shp_T6_multi
        shape_function_matrix_evaluator = shape_function_matrix_T6_multi
        shape_function_derivative_evaluator = dshp_T6_multi
        monomial_evaluator = monoms_T6
        quadrature = {
            "full": Gauss_Legendre_Tri_3a,
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s = symbols("r s", real=True)
            monoms = [1, r, s, r**2, s**2, r * s]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            """
            Returns local coordinates of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array(
                [
                    [-1 / 3, -1 / 3],
                    [2 / 3, -1 / 3],
                    [-1 / 3, 2 / 3],
                    [1 / 6, -1 / 3],
                    [1 / 6, 1 / 6],
                    [-1 / 3, 1 / 6],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            """
            Returns the local coordinates of the center of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array([[0.0, 0.0]], dtype=float)

        @classmethod
        def trimap(cls, subdivide: bool = True) -> ndarray:
            if subdivide:
                return np.array([[0, 3, 5], [3, 1, 4], [5, 4, 2], [5, 3, 4]], dtype=int)
            else:
                return np.array([[0, 1, 2]], dtype=int)

    def to_triangles(self) -> ndarray:
        """
        Returns the topology as triangles.
        """
        return T6_to_T3(None, self.topology().to_numpy())[1]

    @classmethod
    def from_TriMesh(cls, *args, coords=None, topo=None, **kwargs):
        from sigmaepsilon.mesh.data.trimesh import TriMesh

        if len(args) > 0 and isinstance(args[0], TriMesh):
            return T3_to_T6(TriMesh.coords(), TriMesh.topology())
        elif coords is not None and topo is not None:
            return T3_to_T6(coords, topo)
        else:
            raise NotImplementedError



# -*- coding: utf-8 -*-
from typing import Tuple, List

import numpy as np
from numpy import ndarray
from sympy import symbols

from ..geometry import PolyCellGeometry3d
from ..data.polycell import PolyCell
from ..utils.cells.tet4 import (
    shp_TET4_multi,
    dshp_TET4_multi,
    shape_function_matrix_TET4_multi,
    monoms_TET4,
)
from ..utils.numint import Gauss_Legendre_Tet_1
from ..utils.tet import vol_tet_bulk
from ..utils.utils import cells_coords


class TET4(PolyCell):
    """
    4-node isoparametric hexahedron.
    """

    label = "TET4"

    class Geometry(PolyCellGeometry3d):
        number_of_nodes = 4
        vtk_cell_id = 10
        shape_function_evaluator = shp_TET4_multi
        shape_function_matrix_evaluator = shape_function_matrix_TET4_multi
        shape_function_derivative_evaluator = dshp_TET4_multi
        monomial_evaluator = monoms_TET4
        quadrature = {
            "full": Gauss_Legendre_Tet_1,
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s, t = symbols("r s t", real=True)
            monoms = [1, r, s, t]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            return np.array(
                [
                    [-1 / 3, -1 / 3, -1 / 3],
                    [2 / 3, -1 / 3, -1 / 3],
                    [-1 / 3, 2 / 3, -1 / 3],
                    [-1 / 3, -1 / 3, 2 / 3],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            return np.array([[0.0, 0.0, 0.0]], dtype=float)

        @classmethod
        def tetmap(cls) -> ndarray:
            return np.array([[0, 1, 2, 3]], dtype=int)

    def to_tetrahedra(self, flatten: bool = True) -> ndarray:
        tetra = self.topology().to_numpy()
        if flatten:
            return tetra
        else:
            return tetra.reshape(len(tetra), 1, 4)

    def volumes(self) -> ndarray:
        coords = self.source_coords()
        topo = self.topology().to_numpy()
        volumes = vol_tet_bulk(cells_coords(coords, topo))
        res = np.sum(
            volumes.reshape(topo.shape[0], int(len(volumes) / topo.shape[0])),
            axis=1,
        )
        return res



from numba import njit, prange
import numpy as np
from numpy import ndarray

from sigmaepsilon.math.linalg import inv

__cache = True


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def cell_measures(ecoords: ndarray, dshp: ndarray, qweight: ndarray) -> ndarray:
    nE = ecoords.shape[0]
    res = np.zeros(nE, dtype=ecoords.dtype)
    nQ = qweight.shape[0]
    for iQ in range(nQ):
        _dshp = dshp[iQ]
        for iE in prange(nE):
            jac = ecoords[iE].T @ _dshp
            djac = np.linalg.det(jac)
            res[iE] += qweight[iQ] * djac
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _loc_to_glob_bulk_(shp: ndarray, gcoords: ndarray) -> ndarray:
    """
    Local to global transformation for several cells and points.

    Parameters
    ----------
    shp: numpy.ndarray
        The shape functions evaluated at a 'nP' number of local coordinates.
    gcoords: numpy.ndarray
        3D array of shape (nE, nNE, nD) containing coordinates global for every node of
        a single element.

    Returns
    -------
    numpy.ndarray
        Array of global cooridnates of shape (nE, nP, nD).
    """
    nP = shp.shape[0]
    nE, _, nD = gcoords.shape
    res = np.zeros((nE, nP, nD), shp.dtype)
    for i in prange(nE):
        res[i, :, :] = shp @ gcoords[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _glob_to_loc_bulk_(
    lcoords: ndarray, monoms_glob_cells: ndarray, monoms_glob_points: ndarray
) -> ndarray:
    nP = monoms_glob_points.shape[0]
    nE = monoms_glob_cells.shape[0]
    nD = lcoords.shape[-1]
    res = np.zeros((nE, nP, nD), dtype=lcoords.dtype)
    for i in prange(nE):
        coeffs = inv(monoms_glob_cells[i]).T
        for k in prange(nP):
            shp = coeffs @ monoms_glob_points[k]
            res[i, k, :] = shp @ lcoords
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _glob_to_loc_bulk_2_(
    lcoords: ndarray, monoms_glob_cells: ndarray, monoms_glob_points: ndarray
) -> ndarray:
    nP = monoms_glob_points.shape[0]
    nD = lcoords.shape[-1]
    res = np.zeros((nP, nD), dtype=lcoords.dtype)
    for i in prange(nP):
        coeffs = inv(monoms_glob_cells[i]).T
        shp = coeffs @ monoms_glob_points[i]
        res[i, :] = shp @ lcoords
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _ntet_to_loc_bulk_(
    lcoords: ndarray,
    nat_tet: ndarray,
    tetmap: ndarray,
    cell_tet_indices: ndarray,
    points_to_neighbours: ndarray,
) -> ndarray:
    nP = points_to_neighbours.shape[0]
    res = np.zeros((nP, 3), dtype=lcoords.dtype)
    for i in prange(nP):
        nat = nat_tet[i, points_to_neighbours[i]]
        subtopo = tetmap[cell_tet_indices[i]]
        for j in range(4):
            res[i] += lcoords[subtopo[j]] * nat[j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _ntri_to_loc_bulk_(
    lcoords: ndarray,
    nat_tri: ndarray,
    trimap: ndarray,
    cell_tri_indices: ndarray,
    points_to_neighbours: ndarray,
) -> ndarray:
    nP = points_to_neighbours.shape[0]
    res = np.zeros((nP, 2), dtype=lcoords.dtype)
    for i in prange(nP):
        nat = nat_tri[i, points_to_neighbours[i]]
        subtopo = trimap[cell_tri_indices[i]]
        for j in range(3):
            res[i] += lcoords[subtopo[j]] * nat[j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _find_first_hits_1d_(indices: ndarray) -> ndarray:
    N = np.max(indices) + 1
    res = np.zeros(N, dtype=indices.dtype)
    for i in prange(indices.shape[0]):
        res[indices[i]] = i
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _find_first_hits_(pips: ndarray) -> ndarray:
    nP, nE = pips.shape
    global_indices = np.zeros(nP, dtype=np.int64)
    relative_indices = np.zeros(nP, dtype=np.int64)
    for i in prange(nP):
        for j in prange(nE):
            if pips[i, j]:
                global_indices[i] = j
                relative_indices[i] = j
    return global_indices, relative_indices


@njit(nogil=True, parallel=True, cache=__cache)
def _find_first_hits_knn_(pips: ndarray, neighbours: ndarray) -> ndarray:
    nP, nE = pips.shape
    global_indices = np.zeros(nP, dtype=np.int64)
    relative_indices = np.zeros(nP, dtype=np.int64)
    for i in prange(nP):
        for j in prange(nE):
            if pips[i, j]:
                global_indices[i] = neighbours[i, j]
                relative_indices[i] = j
    return global_indices, relative_indices



"""
Algorithms to evaluate cell-in-cell calculations. 
"""

import numpy as np
from numpy import ndarray
from numba import njit, prange

from .tet import pip_tet

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def H8_in_TET4_bulk_knn(
    ecoords_H8: ndarray,
    ecoords_TET4: ndarray,
    neighbours: ndarray,
    tol: float = 1e-12,
) -> ndarray:
    """
    Tells if H8 cells are inside TET4 cells. The function returns
    a boolean NumPy array with the same length as the number of H8 cells.
    If the i-th value is `True`, it means that the i-th H8 cell is inside
    at least one TET4 cell.

    The function goes through all nodes of all H8 cells and check if it
    is inside any of the TET4 cells that make up the neighbourhood of the
    cell the point belongs to.

    Parameters
    ----------
    ecoords_H8: numpy.ndarray
        3d NumPy array of the coordinates of the nodes of the H8 cells.
    ecoords_TET4: numpy.ndarray
        3d NumPy array of the coordinates of the nodes of the TET4 cells.
    neighbours: numpy.ndarray
        2d NumPy array of the indices of the neighbours of each H8 cell. These
        indices refer to the TET4 cells in the `ecoords_TET4` array.
    tol: float, Optional
        Tolerance to consider a point inside a cell. Default is 1e-12.
    """
    nN = neighbours.shape[1]
    res = np.zeros(ecoords_H8.shape[0], dtype=np.bool_)
    for i in prange(ecoords_H8.shape[0]):
        pip = False
        for j in range(8):
            p = ecoords_H8[i, j]
            for k in range(nN):
                pip = pip_tet(p, ecoords_TET4[neighbours[i, k]], tol)
                if pip:
                    res[i] = True
                    break
            if pip:
                break
    return res



from typing import Iterable, Union, Optional, Any

import numpy as np
from numpy import ndarray
import awkward as ak
from awkward import Array as akArray, Record as akRecord

from sigmaepsilon.core.wrapping import Wrapper
from sigmaepsilon.core.typing import issequence

AwkwardLike = Union[akArray, akRecord]


__all__ = ["AkWrapper"]


class AkWrapper(Wrapper):
    """
    A wrapper for Awkward objects. This is the base class of many
    database classes in SigmaEpsilon projects.
    """

    _attr_map_ = {}

    def __init__(
        self,
        *args,
        wrap: Optional[Union[Any, None]] = None,
        fields: Optional[Union[Iterable[str], None]],
        **kwargs,
    ):
        fields = {} if fields is None else fields
        assert isinstance(fields, dict)

        if wrap is None and (len(kwargs) + len(fields)) > 0:
            for k, v in kwargs.items():
                if isinstance(v, np.ndarray):
                    fields[k] = v

            if len(fields) > 0:
                wrap = ak.zip(fields, depth_limit=1)

        if len(kwargs) > 0:
            [kwargs.pop(k, None) for k in fields.keys()]

        super().__init__(*args, wrap=wrap, **kwargs)

    @property
    def db(self) -> akRecord:
        """
        Returns the wrapped Awkward object.
        """
        return self._wrapped

    def to_numpy(self, key: str) -> ndarray:
        """
        Returns a data with the specified key as a numpy array, if
        possible.
        """
        return self._wrapped[key].to_numpy()

    def to_dataframe(
        self, *args, fields: Optional[Union[Iterable[str], None]] = None, **kwargs
    ):
        """
        Returns the data of the database as a DataFrame.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        fields: Iterable[str], Optional
            Valid field names to include in the parquet files.
        **kwargs: dict, Optional
            Keyword arguments forwarded to :func:`awkward.to_dataframe`.

        Returns
        -------
        pandas.DataFrame
        """
        akdb = self.to_ak(*args, fields=fields)
        return ak.to_dataframe(akdb, **kwargs)

    def to_parquet(
        self,
        path: str,
        *args,
        fields: Optional[Union[Iterable[str], None]] = None,
        **kwargs,
    ) -> None:
        """
        Saves the data of the database to a parquet file.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        path: str
            Path of the file being created.
        fields: Iterable[str], Optional
            Valid field names to include in the parquet files.
        **kwargs: dict, Optional
            Keyword arguments forwarded to :func:`awkward.to_parquet`.
        """
        if fields is None and len(args) == 0:
            ak.to_parquet(self.db, path, **kwargs)
        else:
            akdb = self.to_ak(*args, fields=fields, asarray=False)
            ak.to_parquet(akdb, path, **kwargs)

    @classmethod
    def from_parquet(cls, path: str) -> "AkWrapper":
        """
        Saves the data of the database to a parquet file.

        Parameters
        ----------
        path: str
            Path of the file being created.
        """
        return cls(wrap=ak.from_parquet(path))

    def to_ak(
        self,
        *args,
        fields: Optional[Union[Iterable[str], None]] = None,
        asarray: Optional[bool] = False,
    ) -> Union[akArray, akRecord]:
        """
        Returns the database with a specified set of fields as either
        an Awkward Record, or an Awkward Array. If there are no fields
        specified and the output is a record, the original database is
        returned.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        fields: Iterable[str], Optional
            Valid field names to include in the returned objects.
        asarray: bool, Optional
            If True, the database is turned onto an Awkward Array before
            saving to file. Default is False.
        """
        if asarray:
            return self.to_akarray(*args, fields=fields)
        else:
            return self.to_akrecord(*args, fields=fields)

    def to_akarray(
        self, *args, fields: Optional[Union[Iterable[str], None]] = None
    ) -> akArray:
        """
        Returns the data of the mesh as an Awkward array.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        fields: Iterable[str], Optional
            Valid field names to include in the returned objects.
        """
        ldb = self.to_list(*args, fields=fields)
        return ak.from_iter(ldb)

    def to_akrecord(
        self, *args, fields: Optional[Union[Iterable[str], None]] = None
    ) -> akRecord:
        """
        Returns the data of the mesh as an Awkward record.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        fields: Iterable[str], Optional
            Valid field names to include in the returned objects.
        """
        d = self.to_dict(*args, fields=fields)
        return ak.zip(d, depth_limit=1)

    def to_dict(
        self, *args, fields: Optional[Union[Iterable[str], None]] = None
    ) -> dict:
        """
        Returns data of the object as a dictionary. Unless fields
        are specified, all fields are returned.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        fields: Iterable[str], Optional
            A list of keys that might identify data in a database.
            Default is None.
        """
        db = self.db
        res = None

        if fields is None:
            fields = []

        fields.extend(args)

        if len(fields) == 0:
            fields = db.fields

        res = {}

        for f in fields:
            if f in db.fields:
                res[f] = db[f]
            else:
                raise KeyError(f"Field {f} not found.")

        return res

    def to_list(
        self, *args, fields: Optional[Union[Iterable[str], None]] = None
    ) -> list:
        """
        Returns data of the object as lists. Unless fields are
        specified, all fields are returned.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        fields: Iterable[str], Optional
            A list of keys that might identify data in a database.
            Default is None.
        """
        db = self.db
        res = None

        if fields is None:
            if len(args) > 0:
                fields = []

        if isinstance(fields, Iterable):
            fields.extend(args)
            db_ = {}
            for f in fields:
                if f in db.fields:
                    db_[f] = db[f]
                else:
                    raise KeyError(f"Field {f} not found.")

            res = AkWrapper(fields=db_).to_list()
        else:
            res = db.to_list()
        return res

    def __len__(self) -> int:
        return len(self._wrapped)

    def __hasattr__(self, attr):
        if attr in self.__class__._attr_map_:
            attr = self.__class__._attr_map_[attr]
        return any([attr in self.__dict__, attr in self._wrapped.__dict__])

    def __getattr__(self, attr):
        if attr in self.__class__._attr_map_:
            attr = self.__class__._attr_map_[attr]

        if attr in self.__dict__:
            return self.__dict__[attr]

        try:
            return getattr(self._wrapped, attr)
        except Exception:
            name = self.__class__.__name__
            raise AttributeError(f"'{name}' object has no attribute called '{attr}'")

    def __getitem__(self, index: str) -> Any:
        is_str = isinstance(index, str)

        if is_str and index in self.__class__._attr_map_:
            index = self.__class__._attr_map_[index]

        return self.db[index]

    def __setitem__(self, index: str, value: Iterable[Any]) -> None:
        if not isinstance(index, str):
            raise TypeError(f"Expected a string, got {type(index)}")

        if not issequence(value):
            raise TypeError(f"Expected a sequence, got {type(value)}")

        if not len(value) == len(self):
            raise ValueError(
                "The provided value must have the same length as the database."
            )

        self._wrapped[index] = value

    def __contains__(self, item: str) -> bool:
        if not isinstance(item, str):
            return False

        if item in self._wrapped.fields:
            return True

        if item in self.__class__._attr_map_:
            return self.__class__._attr_map_[item] in self._wrapped.fields

        return False



# !TODO  handle decimation in all transformations, template : T6_to_T3
import numpy as np
from awkward import Array
from numba import njit, prange
from typing import Union, Sequence, Tuple

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from numpy import ndarray, newaxis
from concurrent.futures import ThreadPoolExecutor

from ..tri import edges_tri
from ..utils import cells_coords
from ..topodata import (
    edgeIds_T3,
    edgeIds_TET4,
    edgeIds_H8,
    edges_Q4,
    edges_H8,
    faces_H8,
    edges_TET4,
    edges_W6,
    faces_W6,
)
from .topo import unique_topo_data
from .trimap import trimap_Q8

__cache = True


__all__ = [
    "transform_topology",
    "compose_trmap",
    "to_T3",
    "L2_to_L3",
    "T3_to_T6",
    "T3_to_L2",
    "T6_to_T3",
    "Q4_to_Q8",
    "Q4_to_Q9",
    "Q9_to_Q4",
    "Q8_to_T3",
    "Q9_to_T6",
    "Q4_to_T3",
    "H8_to_L2",
    "H8_to_Q4",
    "H8_to_H27",
    "H8_to_TET4",
    "H27_to_H8",
    "H27_to_TET10",
    "TET4_to_L2",
    "TET4_to_TET10",
    "TET10_to_TET4",
    "W6_to_W18",
    "W6_to_TET4",
    "W18_to_W6",
]


DataLike = Union[ndarray, Sequence[ndarray]]


def transform_topology(
    topo: ndarray,
    path: ndarray,
    data: ndarray | None = None,
    *_,
    MT: bool = True,
    max_workers: int = 4,
    **__,
) -> Union[ndarray, Tuple[ndarray]]:
    """
    Transforms a mesh by canging the celltype, but keeping the topology of
    the overall configuration.

    Parameters
    ----------
    topo: numpy.ndarray
        The input topology.
    path: numpy.ndarray
        The transformation path.
    data: numpy.ndarray, Optional
        Data attached to the points or the cells. Default is None.
    MT: bool, Optional
        If True, transformation is carried out using multiple threads. Default is True.
    max_workers: int, Optional
        The maximum number of worker threads. Only relevant if MT is True. Default is 4.

    Returns
    -------
    Union[numpy.ndarray, Tuple[numpy.ndarray]]
        Topology, and optionally data as NumPy arrays.

    Example
    -------
    For instance, this could be a way of producing a more coarse triangulation:

    >>> from sigmaepsilon.mesh.utils.topology import transform_topology
    >>> from sigmaepsilon.mesh.triang import triangulate
    >>> from sigmaepsilon.mesh.grid import grid
    >>> coords, topo, _ = triangulate(size=(1, 1), shape=(5, 10))
    >>> nE1 = topo.shape[0]
    >>> coords, topo = T3_to_T6(coords, topo)
    >>> coords, topo = T6_to_T3(coords, topo)
    >>> nE2 = topo.shape[0]
    >>> nE1 * 4 == nE2
    True

    The functions `T3_to_T6` and `T6_to_T3` call `transform_topology` to do
    the transformations. The same with calling `transform_topology` directly
    could be like this:

    >>> coords, topo, _ = triangulate(size=(1, 1), shape=(5, 10))
    >>> nE1 = topo.shape[0]
    >>> coords, topo = T3_to_T6(coords, topo)
    >>> path = np.array(
    ...     [[0, 3, 5], [3, 1, 4], [5, 4, 2], [5, 3, 4]], dtype=int
    ...            )
    >>> topo = transform_topology(topo, path)
    >>> nE2 = topo.shape[0]
    >>> nE1 * 4 == nE2
    True

    In both cases, the original triangulation is splitted to 4 subtriangles per
    each cell. Hence, `transform_topology` is more general, since you have more
    control over how you want to split the cells that make up the original mesh.
    """
    nD = len(path.shape)

    if nD == 1:
        path = path.reshape(1, len(path))

    assert nD <= 2, "Path must be 1 or 2 dimensional."

    if data is None:
        return _transform_topology_(topo, path)
    else:
        if isinstance(data, Array):
            try:
                data = data.to_numpy()
            except Exception:
                raise TypeError("Invalid data type '{}'".format(data.__class__))
        if isinstance(data, ndarray):
            data = transform_topology_data(topo, data, path)
            return _transform_topology_(topo, path), data
        elif isinstance(data, Iterable):

            def foo(d):
                return transform_topology_data(topo, d, path)

            if MT:
                with ThreadPoolExecutor(max_workers=max_workers) as executor:
                    dmap = executor.map(foo, data)
            else:
                dmap = map(foo, data)
            return _transform_topology_(topo, path), list(dmap)


def transform_topology_data(topo: ndarray, data: ndarray, path: ndarray) -> ndarray:
    if data.shape[:2] == topo.shape[:2]:
        # it is assumed that values are provided for each node of each cell
        res = repeat_cell_nodal_data(data, path)
    elif data.shape[0] == topo.shape[0]:
        # assume that data is constant over the elements
        res = np.repeat(data, path.shape[0], axis=0)
    else:
        raise NotImplementedError("Invalid data shape {}".format(data.shape))
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _transform_topology_(topo: ndarray, path: ndarray) -> ndarray:
    nE = len(topo)
    nSub, nSubN = path.shape
    res = np.zeros((nSub * nE, nSubN), dtype=topo.dtype)
    for iE in prange(nE):
        c = iE * nSub
        for jE in prange(nSubN):
            for kE in prange(nSub):
                res[c + kE, jE] = topo[iE, path[kE, jE]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def repeat_cell_nodal_data(edata: ndarray, path: ndarray) -> ndarray:
    nSub, nSubN = path.shape
    nE = edata.shape[0]
    res = np.zeros((nSub * nE, nSubN) + edata.shape[2:], dtype=edata.dtype)
    for i in prange(nE):
        ii = nSub * i
        for j in prange(nSub):
            jj = ii + j
            for k in prange(nSubN):
                res[jj, k] = edata[i, path[j, k]]
    return res


def compose_trmap(map1: ndarray, map2: ndarray) -> ndarray:
    nE1 = map1.shape[0]
    nE2, nNE2 = map2.shape
    res = np.zeros((int(nE1 * nE2), nNE2), dtype=map1.dtype)
    c = 0
    for i in range(nE1):
        for j in range(nE2):
            for k in range(nNE2):
                res[c, k] = map1[i, map2[j, k]]
            c += 1
    return res


def to_T3(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if path is None:
        raise TypeError("Expected Iterable for argument 'path', got 'NoneType'.")

    if not isinstance(path, ndarray):
        raise TypeError(f"Expected 'ndarray' for argument 'path', got {type(path)}.")

    if not len(path.shape) == 2:
        raise ValueError("'path' must be a 2d NumPy array")

    if not path.shape[1] == 3:
        raise ValueError("Invalid 'path'.")

    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def Q8_to_T3(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if path is None:
        path = trimap_Q8()
    elif isinstance(path, str):
        raise NotImplementedError
    return to_T3(coords, topo, data, path=path, **kwargs)


def T6_to_T3(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    subdivide: bool = True,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 3
    else:
        if path is None:
            if subdivide:
                path = np.array(
                    [[0, 3, 5], [3, 1, 4], [5, 4, 2], [5, 3, 4]], dtype=topo.dtype
                )
            else:
                path = np.array([[0, 1, 2]], dtype=topo.dtype)
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def H27_to_H8(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    subdivide: bool = True,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 8
    else:
        if path is None:
            if subdivide:
                path = np.array(
                    [
                        [0, 8, 24, 11, 16, 22, 26, 20],
                        [8, 1, 9, 24, 22, 17, 21, 26],
                        [24, 9, 2, 10, 26, 21, 18, 23],
                        [11, 24, 10, 3, 20, 26, 23, 19],
                        [16, 22, 26, 20, 4, 12, 25, 15],
                        [22, 17, 21, 26, 12, 5, 13, 25],
                        [26, 21, 18, 23, 25, 13, 6, 14],
                        [20, 26, 23, 19, 15, 25, 14, 7],
                    ],
                    dtype=topo.dtype,
                )
            else:
                path = np.array([[0, 1, 2, 3, 4, 5, 6, 7]], dtype=topo.dtype)
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def Q9_to_Q4(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 4
    else:
        if path is None:
            path = np.array(
                [[0, 4, 8, 7], [4, 1, 5, 8], [8, 5, 2, 6], [7, 8, 6, 3]],
                dtype=topo.dtype,
            )
        elif isinstance(path, str):
            if path == "grid":
                path = np.array(
                    [[0, 3, 4, 1], [3, 6, 7, 4], [4, 7, 8, 5], [1, 4, 5, 2]],
                    dtype=topo.dtype,
                )
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def Q9_to_T6(coords: ndarray, topo: ndarray, path: ndarray = None):
    if path is None:
        path = np.array([[0, 8, 2, 4, 5, 1], [0, 6, 8, 3, 7, 4]], dtype=topo.dtype)
    return _Q9_to_T6(coords, topo, path)


@njit(nogil=True, parallel=True, cache=__cache)
def _Q9_to_T6(coords: ndarray, topo: ndarray, path: ndarray):
    nE = len(topo)
    topoT6 = np.zeros((2 * nE, 6), dtype=topo.dtype)
    for iE in prange(nE):
        c = iE * 2
        for jE in prange(6):
            topoT6[c, jE] = topo[iE, path[0, jE]]
            topoT6[c + 1, jE] = topo[iE, path[1, jE]]
    return coords, topoT6


def H8_to_TET4(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 4
    else:
        if path is None:
            path = np.array(
                [[1, 2, 0, 5], [3, 0, 2, 7], [5, 4, 7, 0], [6, 5, 7, 2], [0, 2, 7, 5]],
                dtype=topo.dtype,
            )
        elif isinstance(path, str):
            raise NotImplementedError
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def H27_to_TET10(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 10
    else:
        if path is None:
            path = np.array(
                [
                    [1, 2, 0, 5, 9, 24, 8, 17, 21, 22],
                    [3, 0, 2, 7, 11, 24, 10, 19, 20, 23],
                    [5, 4, 7, 0, 12, 15, 25, 22, 16, 20],
                    [6, 5, 7, 2, 13, 25, 14, 18, 21, 23],
                    [0, 2, 7, 5, 24, 23, 20, 22, 21, 25],
                ],
                dtype=topo.dtype,
            )
        elif isinstance(path, str):
            raise NotImplementedError
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def H8_to_Q4(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 4
    else:
        if path is None:
            path = np.array(
                [
                    [0, 4, 7, 3],
                    [1, 2, 6, 5],
                    [0, 1, 5, 4],
                    [2, 3, 7, 6],
                    [0, 3, 2, 1],
                    [4, 5, 6, 7],
                ],
                dtype=topo.dtype,
            )
        elif isinstance(path, str):
            raise NotImplementedError
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def T3_to_L2(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[0] == 3, "Invalid shape!"
        assert path.shape[1] == 2, "Invalid shape!"
    else:
        if path is None:
            path = edgeIds_T3()
        else:
            raise NotImplementedError("Invalid path!")
    if data is None:
        nE = len(topo)
        nSub, nSubN = path.shape
        topo = np.reshape(transform_topology(topo, path), (nE, nSub, nSubN))
        edges, _ = unique_topo_data(topo)
        return coords, edges
    else:
        raise NotImplementedError("Data conversion is not available here!")


def TET4_to_L2(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[0] == 6, "Invalid shape!"
        assert path.shape[1] == 2, "Invalid shape!"
    else:
        if path is None:
            path = edgeIds_TET4()
        else:
            raise NotImplementedError("Invalid path!")
    if data is None:
        nE = len(topo)
        nSub, nSubN = path.shape
        topo = np.reshape(transform_topology(topo, path), (nE, nSub, nSubN))
        edges, _ = unique_topo_data(topo)
        return coords, edges
    else:
        raise NotImplementedError("Data conversion is not available here!")


def H8_to_L2(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[0] == 12, "Invalid shape!"
        assert path.shape[1] == 2, "Invalid shape!"
    else:
        if path is None:
            path = edgeIds_H8()
        else:
            raise NotImplementedError("Invalid path!")
    if data is None:
        nE = len(topo)
        nSub, nSubN = path.shape
        topo = np.reshape(transform_topology(topo, path), (nE, nSub, nSubN))
        edges, _ = unique_topo_data(topo)
        return coords, edges
    else:
        raise NotImplementedError("Data conversion is not available here!")


def Q4_to_T3(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 3
    else:
        if path is None:
            path = np.array([[0, 1, 2], [0, 2, 3]], dtype=topo.dtype)
        elif isinstance(path, str):
            if path == "grid":
                path = np.array([[0, 2, 3], [0, 3, 1]], dtype=topo.dtype)
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def L2_to_L3(coords: ndarray, topo: ndarray, order="ikj") -> Tuple[ndarray]:
    nP = len(coords)
    nodes, nodeIDs = unique_topo_data(topo[:, newaxis, :])
    new_coords = np.mean(coords[nodes], axis=1)
    new_topo = nodeIDs + nP
    topo = np.hstack((topo, new_topo))
    coords = np.vstack((coords, new_coords))
    if order == "ikj":
        _buf = np.copy(topo[:, 1])
        topo[:, 1] = topo[:, 2]
        topo[:, 2] = _buf
    else:
        raise NotImplementedError
    return coords, topo


def T3_to_T6(coords: ndarray, topo: ndarray) -> Tuple[ndarray]:
    nP = len(coords)
    edges, edgeIDs = unique_topo_data(edges_tri(topo))
    new_coords = np.mean(coords[edges], axis=1)
    new_topo = edgeIDs + nP
    topo = np.hstack((topo, new_topo))
    coords = np.vstack((coords, new_coords))
    return coords, topo


def Q4_to_Q8(coords: ndarray, topo: ndarray) -> Tuple[ndarray]:
    nP = len(coords)
    edges, edgeIDs = unique_topo_data(edges_Q4(topo))
    new_coords = np.mean(coords[edges], axis=1)
    new_topo = edgeIDs + nP
    topo_res = np.hstack((topo, new_topo))
    coords_res = np.vstack((coords, new_coords))
    return coords_res, topo_res


def Q4_to_Q9(coords: ndarray, topo: ndarray) -> Tuple[ndarray]:
    nP, nE = len(coords), len(topo)
    # new nodes on the edges
    edges, edgeIDs = unique_topo_data(edges_Q4(topo))
    coords_e = np.mean(coords[edges], axis=1)
    topo_e = edgeIDs + nP
    nP += len(coords_e)
    # new coords at element centers
    ecoords = cells_coords(coords, topo)
    coords_c = np.mean(ecoords, axis=1)
    topo_c = np.reshape(np.arange(nE) + nP, (nE, 1))
    # assemble
    topo_res = np.hstack((topo, topo_e, topo_c))
    coords_res = np.vstack((coords, coords_e, coords_c))
    return coords_res, topo_res


def TET4_to_TET10(coords: ndarray, topo: ndarray) -> Tuple[ndarray]:
    nP = len(coords)
    # new nodes on the edges
    edges, edgeIDs = unique_topo_data(edges_TET4(topo))
    coords_e = np.mean(coords[edges], axis=1)
    topo_e = edgeIDs + nP
    topo_res = np.hstack((topo, topo_e))
    coords_res = np.vstack((coords, coords_e))
    return coords_res, topo_res


def TET10_to_TET4(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    subdivide: bool = True,
    **kwargs,
) -> Tuple[ndarray]:
    if isinstance(path, ndarray):
        assert path.shape[1] == 4
    else:
        if path is None:
            if subdivide:
                path = np.array(
                    [
                        [4, 1, 5, 8],
                        [0, 4, 6, 7],
                        [7, 8, 9, 3],
                        [6, 5, 2, 9],
                        [7, 4, 6, 8],
                        [4, 5, 6, 8],
                        [6, 5, 9, 8],
                        [7, 6, 9, 8],
                    ],
                    dtype=topo.dtype,
                )
            else:
                path = np.array([[0, 1, 2, 3]], dtype=topo.dtype)
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def H8_to_H27(coords: ndarray, topo: ndarray) -> Tuple[ndarray]:
    nP, nE = len(coords), len(topo)
    ecoords = cells_coords(coords, topo)
    # new nodes on the edges
    edges, edgeIDs = unique_topo_data(edges_H8(topo))
    coords_e = np.mean(coords[edges], axis=1)
    topo_e = edgeIDs + nP
    nP += len(coords_e)
    # new nodes on face centers
    faces, faceIDs = unique_topo_data(faces_H8(topo))
    coords_f = np.mean(coords[faces], axis=1)
    topo_f = faceIDs + nP
    nP += len(coords_f)
    # register new nodes in the cell centers
    coords_c = np.mean(ecoords, axis=1)
    topo_c = np.reshape(np.arange(nE) + nP, (nE, 1))
    # assemble
    topo_res = np.hstack((topo, topo_e, topo_f, topo_c))
    coords_res = np.vstack((coords, coords_e, coords_f, coords_c))
    return coords_res, topo_res


def W6_to_TET4(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    **kwargs,
):
    if isinstance(path, ndarray):
        assert path.shape[1] == 4
    else:
        if path is None:
            path = np.array(
                [[0, 1, 2, 4], [3, 5, 4, 2], [2, 5, 0, 4]],
                dtype=topo.dtype,
            )
        elif isinstance(path, str):
            raise NotImplementedError
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def W18_to_W6(
    coords: ndarray,
    topo: ndarray,
    data: DataLike = None,
    *args,
    path: ndarray = None,
    subdivide: bool = True,
    **kwargs,
):
    if isinstance(path, ndarray):
        assert path.shape[1] == 6
    else:
        if path is None:
            if subdivide:
                path = np.array(
                    [
                        [15, 13, 16, 9, 4, 10],
                        [17, 16, 14, 11, 10, 5],
                        [17, 15, 16, 11, 9, 10],
                        [12, 15, 17, 3, 9, 11],
                        [6, 1, 7, 15, 13, 16],
                        [8, 6, 7, 17, 15, 16],
                        [8, 7, 2, 17, 16, 14],
                        [8, 0, 6, 17, 12, 15],
                    ],
                    dtype=topo.dtype,
                )
            else:
                path = np.array([[0, 1, 2, 3, 4, 5]], dtype=topo.dtype)
        elif isinstance(path, str):
            raise NotImplementedError
    if data is None:
        return coords, +transform_topology(topo, path, *args, **kwargs)
    else:
        return (coords,) + transform_topology(topo, path, data, *args, **kwargs)


def W6_to_W18(coords: ndarray, topo: ndarray) -> Tuple[ndarray]:
    nP = len(coords)
    # new nodes on the edges
    edges, edgeIDs = unique_topo_data(edges_W6(topo))
    coords_e = np.mean(coords[edges], axis=1)
    topo_e = edgeIDs + nP
    nP += len(coords_e)
    # new nodes on face centers
    faces, faceIDs = unique_topo_data(faces_W6(topo)[0])
    coords_f = np.mean(coords[faces], axis=1)
    topo_f = faceIDs + nP
    nP += len(coords_f)
    # assemble
    topo_res = np.hstack((topo, topo_e, topo_f))
    coords_res = np.vstack((coords, coords_e, coords_f))
    return coords_res, topo_res



from ...config import __hasmatplotlib__

if not __hasmatplotlib__:  # pragma: no cover

    def triplot_mpl_mesh(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

    def triplot_mpl_hinton(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

    def triplot_mpl_data(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Any, Union, Optional, Iterable

    import numpy as np
    from numpy import ndarray

    from mpl_toolkits.axes_grid1 import make_axes_locatable
    import matplotlib.tri as mpltri
    from matplotlib.figure import Figure, Axes

    from sigmaepsilon.math.utils import minmax
    from sigmaepsilon.mesh.triang import triobj_to_mpl, get_triobj_data, triangulate
    from sigmaepsilon.mesh.utils.tri import offset_tri
    from sigmaepsilon.mesh.utils import cells_coords, explode_mesh_data_bulk

    from .utils import decorate_mpl_ax, triplotter, TriPatchCollection

    @triplotter
    def triplot_mpl_hinton(
        triobj: Any,
        data: ndarray,
        *_,
        fig: Optional[Union[Figure, None]] = None,
        ax: Optional[Union[Axes, Iterable[Axes], None]] = None,
        lw: Optional[float] = 0.5,
        fcolor: Optional[str] = "b",
        ecolor: Optional[str] = "k",
        title: Optional[Union[str, None]] = None,
        suptitle: Optional[Union[str, None]] = None,
        label: Optional[Union[str, None]] = None,
        **kwargs,
    ) -> Any:
        """
        Creates a hinton plot of triangles. The idea is from this example from
        `Matplotlib`:

        https://matplotlib.org/stable/gallery/specialty_plots/hinton_demo.html

        Parameters
        ----------
        triobj: 'TriangulationLike'
            This is either a tuple of mesh data (coordinates and topology)
            or a triangulation object understood by :func:~`sigmaepsilon.mesh.triang.triangulate`.
        data: numpy.ndarray, Optional
            Some data to plot as an 1d NumPy array.
        fig: matplotlib.figure.Figure, Optional
            A `matplotlib` figure to plot on. Default is `None`.
        ax: matplotlib.axes.Axes or Iterable[matplotlib.axes.Axes], Optional.
            A `matplotlib` axis or more if data is 2 dimensional. Default is `None`.
        ecolor: str, Optional
            The color of the edges of the original triangles, before scaling. Default is 'k'.
        fcolor: str, Optional
            The color of the triangles. Default is 'b'.
        lw: Number, Optional
            The linewidth. Default is 0.5.
        title: str, Optional
            Title of the plot. See `matplotlib` for further details. Default is `None`.
        suptitle: str, Optional
            The subtitle of the plot. See `matplotlib` for further details. Default is `None`.
        label: str or Iterable[str], Optional
            The label of the axis. Default is `None`.
        **kwargs**: dict, Optional
            The extra keyword arguments are forwarded to
            :func:`~sigmaepsilon.mesh.plotting.mpl.utils.decorate_mpl_ax`.

        See also
        --------
        :func:`~sigmaepsilon.mesh.plotting.mpl.utils.decorate_mpl_ax`

        Example
        -------
        .. plot::
            :include-source: True

            import numpy as np
            from sigmaepsilon.mesh import grid
            from sigmaepsilon.mesh.utils.topology.tr import Q4_to_T3
            from sigmaepsilon.mesh import triangulate
            from sigmaepsilon.mesh.plotting.mpl.triplot import triplot_mpl_hinton

            gridparams = {
                "size": (1200, 600),
                "shape": (30, 15),
                "eshape": (2, 2),
            }
            coordsQ4, topoQ4 = grid(**gridparams)
            points, triangles = Q4_to_T3(coordsQ4, topoQ4, path="grid")
            triobj = triangulate(points=points[:, :2], triangles=triangles)[-1]

            data = np.random.rand(len(triangles))
            triplot_mpl_hinton(triobj, data=data)
        """
        axobj = None
        tri = triobj_to_mpl(triobj)
        points, triangles = get_triobj_data(tri, trim2d=True)
        cellcoords = offset_tri(points, triangles, data)
        patch = TriPatchCollection(cellcoords, fc=fcolor, ec=ecolor, lw=lw)
        axobj = ax.add_collection(patch)
        decorate_mpl_ax(
            fig=fig,
            ax=ax,
            points=points,
            title=title,
            suptitle=suptitle,
            label=label,
            **kwargs,
        )
        return axobj

    @triplotter
    def triplot_mpl_mesh(
        triobj: Any,
        *_,
        fig: Optional[Union[Figure, None]] = None,
        ax: Optional[Union[Axes, None]] = None,
        lw: Optional[float] = 0.5,
        marker: Optional[str] = "b-",
        zorder: Optional[Union[int, None]] = None,
        fcolor: Optional[Union[str, None]] = None,
        ecolor: Optional[str] = "k",
        title: Optional[Union[str, None]] = None,
        suptitle: Optional[Union[str, None]] = None,
        label: Optional[Union[str, None]] = None,
        **kwargs,
    ) -> Any:
        """
        Plots the mesh of a triangulation.

        Parameters
        ----------
        triobj: 'TriangulationLike'
            This is either a tuple of mesh data (coordinates and topology)
            or a triangulation object understood by :func:~`sigmaepsilon.mesh.triang.triangulate`.
        fig: matplotlib.figure.Figure, Optional
            A `matplotlib` figure to plot on. Default is `None`.
        ax: matplotlib.axes.Axes or Iterable[matplotlib.axes.Axes], Optional.
            A `matplotlib` axis or more if data is 2 dimensional. Default is `None`.
        ecolor: str, Optional
            The color of the edges of the original triangles, before scaling. Default is 'k'.
        fcolor: str, Optional
            The color of the triangles. Default is 'b'.
        lw: Number, Optional
            The linewidth. Default is 0.5.
        zorder: int, Optional
            The zorder of the plot. See `matplotlib` for further details. Default is `None`.
        title: str, Optional
            Title of the plot. See `matplotlib` for further details. Default is `None`.
        suptitle: str, Optional
            The subtitle of the plot. See `matplotlib` for further details. Default is `None`.
        label: str or Iterable[str], Optional
            The label of the axis. Default is `None`.
        **kwargs**: dict, Optional
            The extra keyword arguments are forwarded to
            :func:`~sigmaepsilon.mesh.plotting.mpl.utils.decorate_mpl_ax`.

        See also
        --------
        :func:`~sigmaepsilon.mesh.plotting.mpl.utils.decorate_mpl_ax`

        Example
        -------
        .. plot::
            :include-source: True

            from sigmaepsilon.mesh import grid
            from sigmaepsilon.mesh.utils.topology.tr import Q4_to_T3
            from sigmaepsilon.mesh import triangulate
            from sigmaepsilon.mesh.plotting.mpl.triplot import triplot_mpl_mesh

            gridparams = {
                "size": (1200, 600),
                "shape": (30, 15),
                "eshape": (2, 2),
            }
            coordsQ4, topoQ4 = grid(**gridparams)
            points, triangles = Q4_to_T3(coordsQ4, topoQ4, path="grid")
            triobj = triangulate(points=points[:, :2], triangles=triangles)[-1]

            triplot_mpl_mesh(triobj)
        """
        axobj = None
        tri = triobj_to_mpl(triobj)
        points, triangles = get_triobj_data(tri, trim2d=True)

        if fcolor is None:
            if zorder is not None:
                axobj = ax.triplot(tri, marker, lw=lw, zorder=zorder, **kwargs)
            else:
                axobj = ax.triplot(tri, marker, lw=lw, **kwargs)
        else:
            cellcoords = cells_coords(points, triangles)
            axobj = TriPatchCollection(cellcoords, fc=fcolor, ec=ecolor, lw=lw)
            ax.add_collection(axobj)
        decorate_mpl_ax(
            fig=fig,
            ax=ax,
            points=points,
            title=title,
            suptitle=suptitle,
            label=label,
            **kwargs,
        )
        return axobj

    @triplotter
    def triplot_mpl_data(
        triobj: Any,
        data: ndarray,
        *_,
        fig: Optional[Union[Figure, None]] = None,
        ax: Optional[Union[Axes, Iterable[Axes], None]] = None,
        cmap: Optional[str] = "jet",
        ecolor: Optional[str] = "k",
        lw: Optional[float] = 0.1,
        title: Optional[Union[str, None]] = None,
        suptitle: Optional[Union[str, None]] = None,
        label: Optional[Union[str, None]] = None,
        nlevels: Optional[Union[int, None]] = None,
        refine: Optional[bool] = False,
        refiner: Optional[Union[Any, None]] = None,
        colorbar: Optional[bool] = True,
        subdiv: Optional[int] = 3,
        cbpad: Optional[str] = "2%",
        cbsize: Optional[str] = "5%",
        cbpos: Optional[str] = "right",
        draw_contours: Optional[bool] = True,
        shading: Optional[str] = "flat",
        contour_colors: Optional[str] = "auto",
        **kwargs,
    ) -> Any:
        """
        Convenience function to plot data over triangulations using `matplotlib`. Depending on
        the arguments, the function calls `matplotlib.pyplot.tricontourf` (optionally followed
        by `matplotlib.pyplot.tricontour`) or `matplotlib.pyplot.tripcolor`.

        Parameters
        ----------
        triobj: 'TriangulationLike'
            This is either a tuple of mesh data (coordinates and topology)
            or a triangulation object understood by :func:~`sigmaepsilon.mesh.triang.triangulate`.
        data: numpy.ndarray, Optional
            Some data to plot as an 1d or 2d NumPy array.
        fig: matplotlib.figure.Figure, Optional
            A `matplotlib` figure to plot on. Default is `None`.
        ax: matplotlib.axes.Axes or Iterable[matplotlib.axes.Axes], Optional.
            A `matplotlib` axis or more if data is 2 dimensional. Default is `None`.
        cmap: str, Optional
            The name of a colormap. The default is 'jet'.
        ecolor: str, Optional
            The color of the edges of the triangles. This is only used if data
            is provided over the cells. Default is 'k'.
        lw: Number, Optional
            The linewidth. This is only used if data is provided over the cells. Default is 0.1.
        title: str, Optional
            Title of the plot. See `matplotlib` for further details. Default is `None`.
        suptitle: str, Optional
            The subtitle of the plot. See `matplotlib` for further details. Default is `None`.
        label: str or Iterable[str], Optional
            The label of the axis or more labels if data is 2 dimensional and there are more axes.
            See `matplotlib` for further details. Default is `None`.
        nlevels: int, Optional
            Number of levels on the colorbar, only if `colorbar` is `True`. Default is `None`, in which case
            the colorbar has a continuous distribution. See the examples for the details.
        refine: bool, Optional
            Wether to refine the values. Default is `False`.
        refiner: Any, Optional
            A valid `matplotlib` refiner, only if `refine` is True. If not specified, a `UniformTriRefiner`
            is used. Default is `None`.
        subdiv: int, Optional
            Number of subdivisions for the refiner, only if `refine` is `True`. Default is 3.
        cbpad: str, Optional
            The padding of the colorbar. Default is "2%".
        cbpos: str, Optional
            The position of the colorbar. Default is "right".
        colorbar: bool, Optional
            Wether to put a colorbar or not. Default is `False`.
        draw_contours: bool, Optional
            Wether to draw contour levels or not. Only if data is provided over the cells and `nlevels` is
            also specified. Default is `True`.
        contour_colors: str, Optional
            The color of the contourlines, only if `draw_contours` is `True`. Default is 'auto', which means
            the same as is used for the plot.
        shading: str, Optional
            Shading for `matplotlib.pyplot.tripcolor`, for the case if `nlevels` is None. Default is 'flat'.
        **kwargs**: dict, Optional
            The extra keyword arguments are forwarded to
            :func:`~sigmaepsilon.mesh.plotting.mpl.utils.decorate_mpl_ax`.

        See also
        --------
        :func:`~sigmaepsilon.mesh.plotting.mpl.utils.decorate_mpl_ax`

        Examples
        --------
        .. plot::
            :include-source: True

            import numpy as np
            from sigmaepsilon.mesh import grid
            from sigmaepsilon.mesh.utils.topology.tr import Q4_to_T3
            from sigmaepsilon.mesh import triangulate
            from sigmaepsilon.mesh.plotting.mpl.triplot import triplot_mpl_data

            gridparams = {
                "size": (1200, 600),
                "shape": (30, 15),
                "eshape": (2, 2),
            }
            coordsQ4, topoQ4 = grid(**gridparams)
            points, triangles = Q4_to_T3(coordsQ4, topoQ4, path="grid")
            triobj = triangulate(points=points[:, :2], triangles=triangles)[-1]

            # Data defined over the triangles
            data = np.random.rand(len(triangles))
            triplot_mpl_data(triobj, data=data)

            # Data defined over the points
            data = np.random.rand(len(points))
            triplot_mpl_data(
                triobj, data=data, cmap="jet", nlevels=10, refine=True, draw_contours=True
            )
        """

        axobj = None
        tri = triobj_to_mpl(triobj)
        points, triangles = get_triobj_data(tri, trim2d=True)
        dmin, dmax = minmax(data)

        if refiner is not None:
            refine = True

        nData = len(data)
        if nData == len(triangles):
            nD = len(data.shape)
            if nD == 1:
                axobj = ax.tripcolor(
                    tri, facecolors=data, cmap=cmap, edgecolors=ecolor, lw=lw
                )
            elif nD == 2 and data.shape[1] == 3:
                points, triangles, data = explode_mesh_data_bulk(
                    points, triangles, data
                )
                triobj = triangulate(points=points, triangles=triangles)[-1]
                tri = triobj_to_mpl(triobj)
                axobj = ax.tripcolor(tri, data, cmap=cmap, edgecolors=ecolor, lw=lw)
        elif nData == len(points):
            if refine:
                if refiner is None:
                    refiner = mpltri.UniformTriRefiner(triobj)
                tri, data = refiner.refine_field(data, subdiv=subdiv)

            if isinstance(nlevels, int):
                levels = np.linspace(dmin, dmax, nlevels + 1)
                axobj = ax.tricontourf(tri, data, levels=levels, cmap=cmap)
                if draw_contours:
                    contour_colors = (
                        None if contour_colors == "auto" else contour_colors
                    )
                    ax.tricontour(tri, data, levels=levels, colors=contour_colors)
            else:
                axobj = ax.tripcolor(tri, data, cmap=cmap, shading=shading)
                dmin = axobj.get_array().min()
                dmax = axobj.get_array().max()

        assert axobj is not None, "Failed to handle the provided data."
        if colorbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes(cbpos, size=cbsize, pad=cbpad)
            cbar = fig.colorbar(axobj, cax=cax)
            cbar.ax.set_yticks([dmin, dmax])
            cbar.ax.set_yticklabels(["{:4.2f}".format(i) for i in [dmin, dmax]])

        decorate_mpl_ax(
            fig=fig,
            ax=ax,
            points=points,
            title=title,
            suptitle=suptitle,
            label=label,
            **kwargs,
        )
        return axobj


__all__ = ["triplot_mpl_hinton", "triplot_mpl_mesh", "triplot_mpl_data"]



from typing import Tuple


def rgb_to_hex(rgb: Tuple[int]) -> str:
    """
    Converts from RGB to hexadecimal representation of a color.

    Example
    -------
    >>> from sigmaepsilon.mesh.utils.colors import rgb_to_hex
    >>> rgb_to_hex((255, 255, 255))
    'ffffff'
    """
    return "%02x%02x%02x" % rgb


def hex_to_rgb(value: str) -> Tuple[int]:
    """
    Converts from hexadecimal to RGB representation of a color.

    Example
    -------
    >>> from sigmaepsilon.mesh.utils.colors import hex_to_rgb
    >>> hex_to_rgb("FF65BA")
    (255, 101, 186)
    """
    value = value.lstrip("#")
    lv = len(value)
    return tuple(int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3))



from typing import Tuple, List
from functools import partial

import numpy as np
from numpy import ndarray
import sympy as sy

from ..geometry import PolyCellGeometry3d
from ..data.polycell import PolyCell
from ..utils.cells.h27 import (
    shp_H27_multi,
    dshp_H27_multi,
    shape_function_matrix_H27_multi,
    monoms_H27,
)
from ..utils.numint import Gauss_Legendre_Hex_Grid


class H27(PolyCell):
    """
    27-node isoparametric triquadratic hexahedron.

    ::

        top
        7---14---6
        |    |   |
        15--25--13
        |    |   |
        4---12---5

        middle
        19--23--18
        |    |   |
        20--26--21
        |    |   |
        16--22--17

        bottom
        3---10---2
        |    |   |
        11--24---9
        |    |   |
        0----8---1
    """

    label = "H27"

    class Geometry(PolyCellGeometry3d):
        number_of_nodes = 27
        vtk_cell_id = 29
        shape_function_evaluator = shp_H27_multi
        shape_function_matrix_evaluator = shape_function_matrix_H27_multi
        shape_function_derivative_evaluator = dshp_H27_multi
        monomial_evaluator = monoms_H27
        quadrature = {
            "full": partial(Gauss_Legendre_Hex_Grid, 3, 3, 3),
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.

            """
            locvars = r, s, t = sy.symbols("r s t", real=True)
            monoms = [
                1,
                r,
                s,
                t,
                s * t,
                r * t,
                r * s,
                r * s * t,
                r**2,
                s**2,
                t**2,
                r**2 * s,
                r * s**2,
                r * t**2,
                r**2 * t,
                s**2 * t,
                s * t**2,
                r**2 * s * t,
                r * s**2 * t,
                r * s * t**2,
                r**2 * s**2,
                s**2 * t**2,
                r**2 * t**2,
                r**2 * s**2 * t**2,
                r**2 * s**2 * t,
                r**2 * s * t**2,
                r * s**2 * t**2,
            ]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            """
            Returns local coordinates of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array(
                [
                    [-1.0, -1.0, -1],
                    [1.0, -1.0, -1.0],
                    [1.0, 1.0, -1.0],
                    [-1.0, 1.0, -1.0],
                    [-1.0, -1.0, 1.0],
                    [1.0, -1.0, 1.0],
                    [1.0, 1.0, 1.0],
                    [-1.0, 1.0, 1.0],
                    [0.0, -1.0, -1.0],
                    [1.0, 0.0, -1.0],
                    [0.0, 1.0, -1.0],
                    [-1.0, 0.0, -1.0],
                    [0.0, -1.0, 1.0],
                    [1.0, 0.0, 1.0],
                    [0.0, 1.0, 1.0],
                    [-1.0, 0.0, 1.0],
                    [-1.0, -1.0, 0.0],
                    [1.0, -1.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [-1.0, 1.0, 0.0],
                    [-1.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0],
                    [0.0, 1.0, 0.0],
                    [0.0, 0.0, -1.0],
                    [0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            """
            Returns the local coordinates of the center of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array([0.0, 0.0, 0.0])
